@@ -5,7 +5,7 @@
 
 
 DbResult::DbResult(const DbConnectionPtr& pConn, const std::string& sql) :
-pConn_(pConn)
+  pConn_(pConn)
 {
   pConn->check_connection();
   pConn->set_current_result(this);
@@ -27,6 +27,11 @@ DbResult::~DbResult() {
   } catch (...) {}
 }
 
+DbResult* DbResult::create_and_send_query(const DbConnectionPtr& con, const std::string& sql, bool is_statement) {
+  (void)is_statement;
+  return new DbResult(con, sql);
+}
+
 void DbResult::bind(const List& params) {
   return impl->bind(params);
 }
